; UART driver
; vim: set ft=asm:

.equ BAUD = 57600
.equ UBRR = ((F_CPU + BAUD * 8) / (BAUD * 16) - 1)

.cseg

; initialize uart
uart_init:
	push	r16

	ldi	r16, high(UBRR)
	out	UBRRH, r16
	ldi	r16, low(UBRR)
	out	UBRRL, r16

	ldi	r16, 0
	out	UCSRA, r16
	ldi	r16, (1 << RXEN) | (1 << TXEN)
	out	UCSRB, r16
	ldi	r16, (1 << URSEL) | (1 << UCSZ1) | (1 << UCSZ0)
	out	UCSRC, r16

	pop	r16
	ret

; check if input data is available
; r16 = 0x00 (no data) or 0xFF (data available)
uart_has_data:
	in	r16, UCSRA
	andi	r16, (1 << RXC)
	sbrc	r16, RXC
	ldi	r16, 0xFF
	ret

; read raw character to r16, blocking
uart_raw_getc:
	in	r16, UCSRA
	andi	r16, (1 << RXC)
	breq	uart_raw_getc
	in	r16, UDR
	ret

; read character to r16 with echo
; echo '\n' as '\r\n'
uart_getc:
	rcall	uart_raw_getc
	cpi	r16, 13
	brne	uart_getc_done
	subi	r16, 3
	rcall	uart_raw_putc
	subi	r16, -3
uart_getc_done:
	rcall	uart_raw_putc
	ret

; write raw character in r16, blocking
uart_raw_putc:
	push	r17
uart_raw_putc_loop:
	in	r17, UCSRA
	andi	r17, (1 << UDRE)
	breq	uart_raw_putc_loop
	out	UDR, r16
	pop	r17
	ret

; write character in r16
; output '\n' as '\r\n'
uart_putc:
	cpi	r16, 10
	brne	uart_putc_done
	subi	r16, -3
	rcall	uart_raw_putc
	subi	r16, 3
uart_putc_done:
	rcall	uart_raw_putc
	ret

; read 4-bit hexadecimal number to r16
; return 0 on failure
.macro Mgethex4
	rcall	uart_raw_getc
	cpi	r16, '0'	; smaller than '0'
	brlo	fail		; -> fail
	cpi	r16, ':'	; smaller than ':'
	brlo	digit		; -> digit
	andi	r16, ~0x20	; make upper case
	cpi	r16, 'A'	; smaller than 'A'
	brlo	fail		; -> fail
	cpi	r16, 'G'	; smaller than 'G'
	brlo	letter		; -> letter
fail:	ldi	r16, 0
	ret
digit:	subi	r16, 48
	ret
letter:	subi	r16, 55
	ret
.endm

; read 8-bit hexadecimal number to r16
; return 0 on failure
.macro Mgethex8
	rcall	uart_gethex4	; get first digit
	mov	r17, r16
	rcall	uart_gethex4	; get second digit
	swap	r17
	or	r16, r17	; combine digits
	ret
.endm

uart_gethex4:
	Mgethex4

uart_gethex8:
	Mgethex8

; write 4-bit hexadecimal digit in r16
.macro Mputhex4
	push	r16
	andi	r16, 0x0F
	cpi	r16, 10		; smaller than 10
	brlo	digit		; -> digit
	subi	r16, -7		; else letter
digit:	subi	r16, -'0'
	rcall	uart_raw_putc
	pop	r16
	ret
.endm

; write 8-bit (r16) or 16-bit (r17:r16) hexadecimal number
uart_puthex16:
	push	r16
	mov	r16, r17
	rcall	uart_puthex8
	pop	r16
uart_puthex8:
	swap	r16
	rcall	uart_puthex4
	swap	r16
uart_puthex4:
	Mputhex4

; write flash string in Z
.macro Mputstr
	push	r16
	lpm	r16, Z+		; first char
loop:	rcall	uart_putc	; write it
	lpm	r16, Z+		; next char
	cpi	r16, 0		; if not zero
	brne	loop		; -> loop
	pop	r16
	ret
.endm
uart_putstr:
	Mputstr

; write formatted flash string in Z
;   %% - percent sign
;   %x -  8-bit hex
;   %X - 16-bit hex
.macro Mprintf
	pop	r0		; take return address
	pop	r1		; off the stack

loop:	lpm	r16, Z+		; next char
	cpi	r16, 0		; if zero
	breq	done		; -> done
	cpi	r16, '%'	; if percent
	breq	handle		; -> handle
print:	rcall	uart_putc	; else print
	rjmp	loop		; and loop

handle:	lpm	r16, Z+		; specifier
	cpi	r16, '%'	; %%
	breq	print		; print and loop
	cpi	r16, 'x'	; %x
	breq	hex8		; -> 8-bit hex
	cpi	r16, 'X'	; %X
	breq	hex16		; -> 16-bit hex
	ldi	r16, '?'	; else unknown
	rjmp	print		; print and loop

hex16:	pop	r16		; first byte
	rcall	uart_puthex8	; print hex
hex8:	pop	r16		; second byte
	rcall	uart_puthex8	; print hex
	rjmp	loop		; loop

done:
	push	r1		; restore return address
	push	r0		; on stack
	ret
.endm
uart_printf:
	Mprintf
