; i8080 new bios
; avoid Z80 opcodes!

; system configuration
MSIZE:   equ 60    ; memory size

UARTD:   equ 0x10  ; uart data
UARTS:   equ 0x11  ; uart status

SPID:    equ 0x18  ; spi data
SPISEL:  equ 0x19  ; spi chip select
SPISPD:  equ 0x1A  ; spi speed select

; constants
BIAS:    equ (MSIZE-20) * 1024
CCP:     equ 0x3400 + BIAS      ; ccp  base
BDOS:    equ CCP + 0x0806       ; bdos base
BIOS:    equ CCP + 0x1600       ; bios base
NSECTS:  equ (BIOS-CCP) / 128   ; load size

IOBYTE:  equ 3
CDISK:   equ 4

; entry points
org BIOS
          jp boot            ; cold start
          jp wboot           ; warm start
          jp const           ; console status
          jp conin           ; console input
          jp conout          ; console output
          jp list            ; printer output
          jp punch           ; punch output
          jp reader          ; reader input
          jp home            ; move head to home
          jp seldsk          ; select disk
          jp settrk          ; set track number
          jp setsec          ; set sector number
          jp setdma          ; set dma address
          jp read            ; read disk
          jp write           ; write disk
          jp listst          ; printer status
          jp sectran         ; sector translate

; =====================================
; initialization functions

boot:     ; move initialization code to 0x1000
          ; clear original area (used as bss)
          ld   hl, BSSSEG
          ld   de, 0x1000
          ld   bc, INITLEN
boot1:    ld   a, (hl)       ; read src
          ld   (de), a       ; write dst
          xor  a
          ld   (hl), a       ; clear src
          inc  hl
          inc  de
          dec  bc
          ld   a, b
          or   c
          jp   nz, boot1

          ld   sp, STACK
          call init
          call loadcpm
          call prepcpm
          call autosub
          jp   gocpm

; cold boot replacement
; reload bootstrap and execute it
reboot:   di
          ; TODO
          ld   hl, REBOOT
          call outstr
          halt

; warm boot
wboot:    ld   sp, STACK
          call loadcpm
          call prepcpm
          jp   gocpm

; load cp/m from disk
loadcpm:  ld   c,  0         ; first drive
          call seldsk
          call home
          ld   b,  NSECTS    ; sectors to load
          ld   c,  0         ; track number
          ld   de, 0x0400    ; start sector in D
          ld   hl, CCP       ; destination
loadcpm1: push bc
          push de
          push hl
          ld   b,  0
          ld   c,  d
          call setsec        ; set next sector
          pop  bc
          push bc
          call setdma        ; set load addr
          call read
          jp   nz, loadcpm   ; on error, start over
          pop  hl
          ld   de, 128
          add  hl, de        ; dest += 128
          pop  de
          pop  bc
          dec  b             ; one sector less
          ret  z             ; no more sectors, exit
          inc  d
          ld   a,  (BOOTSPT) ; sectors per track
          cp   d             ; end of track?
          jp   nz, loadcpm1  ; ...no, continue
          ld   d,  0         ; first sector
          inc  c             ; next track
          push bc
          push de
          push hl
          call settrk        ; set track
          pop  hl
          pop  de
          pop  bc
          jp   loadcpm1      ; continue loading

; prepare cp/m
prepcpm:  ld   a, 0xC3       ; JP opcode
          ld   (0x0000), a
          ld   hl, BIOS+3    ; warm boot entry
          ld   (0x0001), hl
          ld   (0x0005), a
          ld   hl, BDOS      ; bdos entry
          ld   (0x0006), hl
          call blkres        ; reset (de)blocking
          ld   bc, 0x0080    ; set default addr
          call setdma
          ret

; run cp/m
gocpm:    ld   a, (CDISK)    ; current user/disk
          and  0x0F          ; disk only
          call dskst         ; check status
          jp   z, gocpm1     ; zf set = disk ok
          xor  a             ; invalid, set A:
          ld   (CDISK), a
gocpm1:   ld   a, (CDISK)    ; set disk
          ld   c, a
          jp   CCP           ; start CCP

; copy B bytes from HL to DE
memcpy:   ld   a, (hl)       ; read from src
          inc  hl
          ld   (de), a       ; write to dst
          inc  de
          dec  b             ; done?
          jp   nz, memcpy    ; ...no, loop
          ret

; print string
outstr:   ld   a, (hl)       ; input in (HL)
          or   a
          ret  z
          ld   c, a
          call conout
          inc  hl
          jp   outstr

; print hexadecimal
outhex16: ld   a, h          ; input in HL
          call outhex8
          ld   a, l
outhex8:  push af            ; input in A
          rra
          rra
          rra
          rra
          call outhex4
          pop  af
outhex4:  and  0x0F          ; input in A
          add  a, 0x90
          daa
          adc  a, 0x40
          daa
          ld   c, a
          call conout
          ret

; =====================================
; console functions - always UART

const:    in   a, (UARTS)
          or   a
          ret  z             ; 0x00 -> no data
          ld   a, 0xFF       ; 0xFF -> data
          ret

conin:    call const
          jp   z, conin      ; wait for data
          in   a, (UARTD)    ; read data
          ret

conout:   ld   a, c
          out  (UARTD), a
          ret

; =====================================
; printer functions - not supported

list:     ret                ; ignore data

listst:   xor a              ; return BUSY
          ret

punch:    ret                ; ignore data

reader:   ld   a, 0x1A       ; return EOF
          ret

; =====================================
; disk functions

seldsk:   jp   dsksel        ; select disk

home:     ld   bc, 0x0000    ; track zero
          ; fall-through
settrk:   ld   h, b
          ld   l, c
          ld   (SEKTRK), hl  ; set track
          ret

setsec:   ld   h, b
          ld   l, c
          ld   (SEKSEC), hl  ; set sector
          ret

setdma:   ld   h, b
          ld   l, c
          ld   (DMAAD), hl   ; set address
          ret

sectran:  ld   h, b          ; no skew support
          ld   l, c
          ret

read:     ld   a, 0x01
          ld   (DSKCMD), a   ; read cmd
          jp   blkrw         ; execute

write:    ld   a, 0x02
          ld   (DSKCMD), a   ; write cmd
          ld   a, c
          ld   (WRTYPE), a   ; write type
          jp   blkrw         ; execute

; =====================================
; blocking and blocking routines
; fixed 512 byte sectors

; reset blocking system
; mark buffer invalid but dirty
blkres:   xor  a
          ld   (HSTACT), a
          ret

; compare translation XLT.. to host HST..
blkcmp:   ld   hl, XLT
          ld   de, HST
          ld   b,  XLTSZ
blkcmp1:  ld   a,  (de)
          cp   (hl)
          ret  nz            ; not equal
          inc  hl
          inc  de
          dec  b
          jp   nz, blkcmp1
          ret                ; equal

; copy translation XLT.. -> host HST..
blksav:   ld   hl, XLT       ; copy XLT -> HST
          ld   de, HST
          ld   b,  XLTSZ
          call memcpy

; copy cp/m SEK.. -> translation XLT..
; and translate by 4:1 deblocking factor
blkxlt:   ld   hl, SEK       ; copy SEK -> XLT
          ld   de, XLT
          ld   b,  XLTSZ
          call memcpy

          ; divide XLTSEC by four
          ld   hl, (XLTSEC)
          ld   a, h          ; rotate HL right
          rrca
          ld   h, a
          ld   a, l
          rra
          ld   l, a
          ld   a, h          ; rotate HL right
          rrca
          ld   h, a
          ld   a, l
          rra
          ld   l, a
          ld   a, h          ; clear top two bits
          and  0x3F
          ld   h, a
          ld   (XLTSEC), hl
          ret

; blocked read/write routine
; no write cache
blkrw:    call blkxlt        ; xlat SEK.. -> XLT..
          call blkcmp        ; in buffer?
          jp   z, blkrw1     ; ...yes, skip read
          call blksav        ; save XLT.. -> HST..
          xor  a
          ld   (HSTACT), a   ; invalidate buffer
          call dskread       ; read physical sector
          or   a
          ret  nz            ; disk error

blkrw1:   ld   a, (DSKCMD)   ; get command
          cp   0x02          ; write?
          jp   z, blkrw2     ; ...yes, write
          call blkxtr        ; extract data
          ld   a, 0xFF
          ld   (HSTACT), a   ; buffer valid
          xor  a             ; success
          ret

blkrw2:   call blkins        ; insert data
          call dskwrite      ; write physical sector
          or   a
          ret  nz            ; disk error
          ld   a, 0xFF
          ld   (HSTACT), a   ; buffer valid
          xor  a             ; success
          ret

; prepare for block/deblock operation
;   outputs: HL = dma buffer
;            DE = cp/m sector address in sector buffer
;            B  = 128
blkprep:  ld   a, (SEKSEC)   ; cp/m sector low
          and  0x03          ; sector index (0..3)
          rrca               ; x 64
          rrca
          ld   e, a          ; DE = 64 x index
          ld   d, 0
          ld   hl, DSKBUF
          add  hl, de        ; HL = buf +  64 x index
          add  hl, de        ; HL = buf + 128 x index
          ex   de, hl        ; DE = buf + 128 x index
          ld   hl, (DMAAD)   ; HL = dma buffer
          ld   b, 128        ; B  = sector length
          ret

; insert cp/m sector into sector buffer
blkins:   call blkprep
          jp   memcpy

; extract cp/m sector from sector buffer
blkxtr:   call blkprep
          ex   de, hl
          jp   memcpy

; =====================================
; disk implementation

include "disks.inc"

; =====================================
; begin of initialized data
DATASEG:  equ $
CODELEN:  equ DATASEG-BIOS

; minicpm disk format
DPB:      dw 256        ; SPT      -> 256 sec/trk
          db 5, 31      ; BSH, BLM -> blocksize 4096
          db 1          ; EXM      -> 1 extent
          dw 2047       ; DSM      -> 2048 blocks total
          dw 1023       ; DRM      -> 1024 directory entries
          db 0xFF, 0x00 ; AL0/AL1  -> 8 blocks for directory
          dw 0          ; CKS      -> no checksum
          dw 1          ; OFS      -> 1 reserved tracks

; disk tables
DPHTAB:   dw 0x0000, 0x0000, 0x0000, 0x0000  ; A:
          dw DIRBUF, DPB,    0x0000, ALV0
          dw 0x0000, 0x0000, 0x0000, 0x0000  ; B:
          dw DIRBUF, DPB,    0x0000, ALV1
          dw 0x0000, 0x0000, 0x0000, 0x0000  ; C:
          dw DIRBUF, DPB,    0x0000, ALV2
          dw 0x0000, 0x0000, 0x0000, 0x0000  ; D:
          dw DIRBUF, DPB,    0x0000, ALV3

; autosub data
AUTOCMD:  db AUTOLEN-2
          db "SUBMIT PROFILE", 0
AUTOLEN:  equ $-AUTOCMD
AUTOFCB1: db '?'             ; drive
          db "SUBMIT  COM"   ; filename
          ds 27              ; fill remainder
AUTOFCB2: db '?'             ; drive
          db "PROFILE SUB"   ; filename
          ds 27              ; fill remainder

; messages
REBOOT:   db "\r\nHARD REBOOT NOT IMPLEMENTED\r\n", 0

BANNER:   db "MINICPM v2\r\n", 0
MEM1:     db "  BIOS:", 0
MEM2:     db " LEN:", 0
MEM3:     db " DATA:", 0
MEM4:     db " BSS:", 0
MEM5:     db " INIT:", 0
MEM6:     db " END:", 0
MEM7:     db 13, 10, 0

; =====================================
; begin of zero-initialized data
; note: overlaid with init code in ROM
BSSSEG:   equ $
DATALEN:  equ BSSSEG-DATASEG

; disk variables
DSKCMD:   ds 1        ; disk operation
                      ; 0x01: read, 0x02: write
WRTYPE:   ds 1        ; write type
                      ; 0x00: normal, 0x01: directory, 0x02: new block
DMAAD:    ds 2        ; cp/m memory address

BOOTDISK: ds 1        ; boot disk
BOOTIO:   ds 1        ; boot iobyte
BOOTSPT:  ds 1        ; boot area sec/trk

; cp/m sector information
SEK:
SEKDSK:   ds 1        ; selected disk
SEKTRK:   ds 2        ; selected track
SEKSEC:   ds 2        ; selected sector
SEKDPH:   ds 2        ; selected DPH
SEKACT:   ds 1        ; always true

; translation sector information
XLT:
XLTDSK:   ds 1        ; translation disk
XLTTRK:   ds 2        ; translation track
XLTSEC:   ds 2        ; translation sector
XLTDPH:   ds 2        ; translation DPH
XLTACT:   ds 1        ; always true
XLTSZ:    equ $ - XLT ; block size

; host sector buffer information
HST:
HSTDSK:   ds 1        ; disk in buffer
HSTTRK:   ds 2        ; track in buffer
HSTSEC:   ds 2        ; sector in buffer
HSTDPH:   ds 2        ; current DPH
HSTACT:   ds 1        ; true if buffer is valid

; physical sector buffer
DSKBUF:   ds 512

; max-sized allocation vectors
ALV0:     ds 256
ALV1:     ds 256
ALV2:     ds 256
ALV3:     ds 256

; cp/m directory buffer
DIRBUF:   ds 128

; reserve stack space
          ds 128
STACK:    equ $

BSSEND:   equ $
BSSLEN:   equ BSSEND-BSSSEG

; =====================================
; initialization code
; moved to 0x1000 during cold boot
; note: overlaid with BSS in ROM
seek BSSSEG-BIOS
org  0x1000
INITSEG:  equ $

; cold boot initialization
init:     di                 ; disable interrupts
          ld   hl, reboot    ; replace boot function
          ld   (BIOS+1), hl

          ld   a, (BOOTIO)   ; set initial iobyte
          ld   (IOBYTE), a

          ld   hl, BANNER    ; print banner
          call outstr

          ld   a, 0xFF       ; initialize constants
          ld   (SEKACT), a
          ld   (XLTACT), a
          call blkres        ; reset (de)blocking
          call dskinit       ; initialize disks

          ld   a, (BOOTDISK) ; set boot drive
          ld   (CDISK), a

          ld   hl, MEM1      ; print memory information
          call outstr
          ld   hl, BIOS      ; bios base
          call outhex16
          ld   hl, MEM2
          call outstr
          ld   hl, CODELEN   ; bios length
          call outhex16
          ld   hl, MEM3
          call outstr
          ld   hl, DATALEN   ; data length
          call outhex16
          ld   hl, MEM4
          call outstr
          ld   hl, BSSLEN    ; bss length
          call outhex16
          ld   hl, MEM5
          call outstr
          ld   hl, INITLEN   ; init length
          call outhex16
          ld   hl, MEM6
          call outstr
          ld   hl, BSSEND    ; runtime top
          call outhex16
          ld   hl, MEM7
          call outstr

          ret

; setup autosub command
; if A:SUBMIT.COM and A:PROFILE.SUB exist
autosub:  ld   a, (BOOTDISK) ; get boot disk
          push af
          inc  a             ; make FCB code
          ld   (AUTOFCB1), a ; write to first FCB
          ld   (AUTOFCB2), a ; write to second FCB

          ld   c, 13         ; BDOS: reset disks
          call BDOS
          pop  af

          ld   c, 17         ; BDOS: find first
          ld   de, AUTOFCB1  ; SUBMIT.COM
          call BDOS
          inc  a
          ret  z             ; not found, exit

          ld   c, 17         ; BDOS: find first
          ld   de, AUTOFCB2  ; PROFILE.SUB
          call BDOS
          inc  a
          ret  z             ; not found, exit

          ld   hl, AUTOCMD   ; copy command
          ld   de, CCP+7     ; into ccp
          ld   b,  AUTOLEN   ; command length
          call memcpy
          ret

INITEND:  equ $
INITLEN:  equ INITEND-INITSEG

